/*
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 *
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2014 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.kiml.util.nodespacing;

import java.util.EnumSet;
import java.util.Iterator;

import de.cau.cs.kieler.core.math.KVector;
import de.cau.cs.kieler.core.properties.IProperty;
import de.cau.cs.kieler.core.properties.Property;
import de.cau.cs.kieler.kiml.options.LabelSide;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.NodeLabelPlacement;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.options.PortLabelPlacement;
import de.cau.cs.kieler.kiml.options.SizeConstraint;
import de.cau.cs.kieler.kiml.options.SizeOptions;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.kiml.util.adapters.GraphAdapters.GraphAdapter;
import de.cau.cs.kieler.kiml.util.adapters.GraphAdapters.LabelAdapter;
import de.cau.cs.kieler.kiml.util.adapters.GraphAdapters.NodeAdapter;
import de.cau.cs.kieler.kiml.util.adapters.GraphAdapters.PortAdapter;
import de.cau.cs.kieler.kiml.util.nodespacing.Spacing.Insets;
import de.cau.cs.kieler.kiml.util.nodespacing.Spacing.Margins;

/**
 * Calculates node sizes, places ports, and places node and port labels.
 * 
 * <p><i>Note:</i> Regarding port placement, this processor now does what the old
 * {@code PortPositionProcessor} did and thus replaces it.</p>
 * 
 * <dl>
 *   <dt>Precondition:</dt>
 *     <dd>The graph is layered.</dd>
 *     <dd>Crossing minimization is finished.</dd>
 *     <dd>Port constraints are at least at {@code FIXED_ORDER}.</dd>
 *     <dd>Port lists are properly sorted going clockwise, starting at the leftmost northern port.</dd>
 *   <dt>Postcondition:</dt>
 *     <dd>Port positions are fixed.</dd>
 *     <dd>Port labels are placed.</dd>
 *     <dd>Node labels are placed.</dd>
 *     <dd>Node sizes are set.</dd>
 *   <dt>Slots:</dt>
 *     <dd>Before phase 4.</dd>
 *   <dt>Same-slot dependencies:</dt>
 *     <dd>{@link LabelSideSelector}</dd>
 * </dl>
 * 
 * @see LabelSideSelector
 * @author cds
 * @author uru
 */
public class LabelAndNodeSizeProcessor {
    
    /**
     * Copy of the
     * {@link de.cau.cs.kieler.klay.layered.properties.InternalProperties#PORT_RATIO_OR_POSITION}
     * option. For further information see the documentation found there. 
     * We added this copy here to allow a generic treatment of spacing calculations for graph elements.
     * See the {@link de.cau.cs.kieler.kiml.util.nodespacing} package.
     * [programmatically set]
     */
    public static final IProperty<Double> PORT_RATIO_OR_POSITION = new Property<Double>(
            "portRatioOrPosition", 0.0);
    
    /**
     * The minimal port spacing we use.
     */
    private static final double MIN_PORT_SPACING = 10;

    /*
     * The following variables provide context information for the different phases of the
     * algorithm. The information are usually computed by some phase to be made available to later
     * phases. This would probably be better kept in some kind of a context object, but is kept here
     * for performance reasons to avoid having to create new context objects all the time.
     */

    /**
     * Node insets required by port labels inside the node. This is always set, but not always taken
     * into account to calculate the node size.
     * 
     * <p>
     * <i>Note:</i> This is only valid for the currently processed node!
     * </p>
     */
    private Insets requiredPortLabelSpace = new Insets();

    /**
     * Node insets required by node labels placed inside the node. This is always set, but not
     * always taken into account to calculate the node size.
     * 
     * <p>
     * <i>Note:</i> This is only valid for the currently processed node!
     * </p>
     */
    private Insets requiredNodeLabelSpace = new Insets();

    /**
     * Space required by the node labels if stacked vertically.
     * 
     * <p>
     * <i>Note:</i> This is only valid for the currently processed node!
     * </p>
     */
    private KVector nodeLabelsBoundingBox = new KVector();

    /**
     * Number of ports on the western side. Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private int westPortsCount = 0;

    /**
     * Height of the ports on the western side. If port labels are accounted for, the height
     * includes the relevant port margins too. Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private double westPortsHeight = 0.0;

    /**
     * Number of ports on the eastern side.Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private int eastPortsCount = 0;

    /**
     * Height of the ports on the eastern side. If port labels are accounted for, the height
     * includes the relevant port margins too. Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private double eastPortsHeight = 0.0;

    /**
     * Number of ports on the northern side.Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private int northPortsCount = 0;

    /**
     * Width of the ports on the northern side. If port labels are accounted for, the height
     * includes the relevant port margins too. Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private double northPortsWidth = 0.0;

    /**
     * Number of ports on the southern side.Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private int southPortsCount = 0;

    /**
     * Width of the ports on the southern side. If port labels are accounted for, the height
     * includes the relevant port margins too. Only used if port constraints are not
     * {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}.
     */
    private double southPortsWidth = 0.0;

    /**
     * Resets the fields providing context information to the algorithm.
     */
    private void resetContext() {
        requiredPortLabelSpace.set(0.0, 0.0, 0.0, 0.0);
        requiredNodeLabelSpace.set(0.0, 0.0, 0.0, 0.0);
        
        westPortsCount = 0;
        westPortsHeight = 0.0;
        eastPortsCount = 0;
        eastPortsHeight = 0.0;
        northPortsCount = 0;
        northPortsWidth = 0.0;
        southPortsCount = 0;
        southPortsWidth = 0.0;
    }
       
    /*
     * Entry point
     */
    /**
     * {@inheritDoc}
     */
    public void process(final GraphAdapter<?> layeredGraph) {
        double labelSpacing = layeredGraph.getProperty(LayoutOptions.LABEL_SPACING);

        // Iterate over all the graph's nodes
        for (NodeAdapter<?> node : layeredGraph.getNodes()) {
            /* Note that, upon Miro's request, each phase of the algorithm was given a code name. */
            
            /* PREPARATIONS
             * Reset stuff and remember the node's old size.
             */
            resetContext();
            KVector originalNodeSize = new KVector(node.getSize());
            double portSpacing = Math.max(MIN_PORT_SPACING,
                    node.getProperty(LayoutOptions.PORT_SPACING));
            
            
            /* PHASE 1 (SAD DUCK): PLACE PORT LABELS
             * Port labels are placed and port margins are calculated. We currently only support
             * one label per port.
             */
            PortLabelPlacement labelPlacement = node.getProperty(LayoutOptions.PORT_LABEL_PLACEMENT);
            boolean compoundNodeMode = node.isCompoundNode();
            
            // Place port labels and calculate the margins
            for (PortAdapter<?> port : node.getPorts()) {
                placePortLabels(port, labelPlacement, compoundNodeMode, labelSpacing);
                calculateAndSetPortMargins(port);
            }
            
            // Count ports on each side and calculate how much space they require
            calculatePortInformation(node, node.getProperty(LayoutOptions.SIZE_CONSTRAINT).contains(
                    SizeConstraint.PORT_LABELS));
            
            
            /* PHASE 2 (DYNAMIC DONALD): CALCULATE INSETS
             * We know the sides the ports will be placed at and we know where node labels are to
             * be placed. Calculate the node's insets accordingly. Also compute the amount of space
             * the node labels will need if stacked vertically. Note that we don't have to know the
             * final position of ports and of node labels to calculate all this stuff.
             */
            calculateRequiredPortLabelSpace(node);
            calculateRequiredNodeLabelSpace(node, labelSpacing);
            
            
            /* PHASE 3 (DANGEROUS DUCKLING): RESIZE NODE
             * If the node has labels, the node insets might have to be adjusted to reserve space
             * for them, which is what this phase does.
             */
            resizeNode(node, portSpacing, labelSpacing);
            
            
            /* PHASE 4 (DUCK AND COVER): PLACE PORTS
             * The node is resized, taking all node size constraints into account. The port spacing
             * is not required for port placement since the placement will be based on the node's
             * size (if it is not fixed anyway).
             */
            placePorts(node, originalNodeSize);
            
            
            /* PHASE 5 (HAPPY DUCK): PLACE NODE LABELS
             * With space reserved for the node labels, the labels are placed.
             */
            placeNodeLabels(node, labelSpacing);
            
            
            /* CLEANUP (THANKSGIVING): SET NODE INSETS
             * Set the node insets to include space required for port and node labels. If the labels
             * were not taken into account when calculating the node's size, this may result in
             * insets that, taken together, are larger than the node's actual size.
             */
            Insets nodeInsets = new Insets(node.getInsets());
            nodeInsets.left = requiredNodeLabelSpace.left + requiredPortLabelSpace.left;
            nodeInsets.right = requiredNodeLabelSpace.right + requiredPortLabelSpace.right;
            nodeInsets.top = requiredNodeLabelSpace.top + requiredPortLabelSpace.top;
            nodeInsets.bottom = requiredNodeLabelSpace.bottom + requiredPortLabelSpace.bottom;
            node.setInsets(nodeInsets);
        }
    }
    
    
    ///////////////////////////////////////////////////////////////////////////////
    // PORT LABEL PLACEMENT

    /**
     * Places the label of the given port, if any. We assume that the label is actually part of the
     * given port.
     * 
     * <p><i>Note:</i> We currently only support one label per port.</p>
     * 
     * @param port the port whose labels to place.
     * @param placement the port label placement that applies to the port.
     * @param compoundNodeMode {@code true} if the node contains further nodes in the original graph.
     *                         This influences the inner port label placement.
     * @param labelSpacing spacing between labels and other objects.
     */
    private void placePortLabels(final PortAdapter<?> port, final PortLabelPlacement placement,
            final boolean compoundNodeMode, final double labelSpacing) {
        
        // Get the port's label, if any
        Iterator<LabelAdapter<?>> labelIter = port.getLabels().iterator();
        if (labelIter.hasNext()) {
            // We use different implementations based on whether port labels are to be placed
            // inside or outside the node
            if (placement.equals(PortLabelPlacement.INSIDE)) {
                placePortLabelsInside(port, labelIter.next(), compoundNodeMode, labelSpacing);
            } else if (placement.equals(PortLabelPlacement.OUTSIDE)) {
                placePortLabelsOutside(port, labelIter.next(), labelSpacing);
            }
        }
    }

    /**
     * Places the label of the given port on the inside of the port's node.
     * 
     * @param port the port whose label to place.
     * @param label the label to place.
     * @param compoundNodeMode {@code true} if the node contains further nodes in the original graph. In
     *                         this case, port labels are not placed next to ports, but a little down as
     *                         well to avoid edge-label-crossings if the port has edges connected to the
     *                         node's insides.
     * @param labelSpacing spacing between labels and other objects.
     */
    private void placePortLabelsInside(final PortAdapter<?> port, final LabelAdapter<?> label,
            final boolean compoundNodeMode, final double labelSpacing) {
        
        KVector position = new KVector(port.getPosition());
        switch (port.getSide()) {
        case WEST:
            position.x = port.getSize().x + labelSpacing;
            position.y = compoundNodeMode && port.hasCompoundConnections()
                    ? port.getSize().y + labelSpacing
                    : (port.getSize().y - label.getSize().y) / 2.0;
            break;
        case EAST:
            position.x = -label.getSize().x - labelSpacing;
            position.y = compoundNodeMode && port.hasCompoundConnections()
                    ? port.getSize().y + labelSpacing
                    : (port.getSize().y - label.getSize().y) / 2.0;
            break;
        case NORTH:
            position.x = -label.getSize().x / 2;
            position.y = port.getSize().y + labelSpacing;
            break;
        case SOUTH:
            position.x = -label.getSize().x / 2;
            position.y = -label.getSize().y - labelSpacing;
            break;
        }
        label.setPosition(position);
    }
    
    /**
     * Places the label of the given port on the outside of the port's node.
     * 
     * @param port the port whose label to place.
     * @param label the label to place.
     * @param labelSpacing spacing between labels and other objects.
     */
    private void placePortLabelsOutside(final PortAdapter<?> port, final LabelAdapter<?> label,
            final double labelSpacing) {
        
        KVector position = new KVector(label.getPosition());
        if (label.getSide() == LabelSide.ABOVE) {
            // Place label "above" edges
            switch (port.getSide()) {
            case WEST:
                position.x = -label.getSize().x - labelSpacing;
                position.y = -label.getSize().y - labelSpacing;
                break;
            case EAST:
                position.x = port.getSize().x + labelSpacing;
                position.y = -label.getSize().y - labelSpacing;
                break;
            case NORTH:
                position.x = -label.getSize().x - labelSpacing;
                position.y = -label.getSize().y - labelSpacing;
                break;
            case SOUTH:
                position.x = -label.getSize().x - labelSpacing;
                position.y = port.getSize().y + labelSpacing;
                break;
            }
        } else {
            // Place label "below" edges
            switch (port.getSide()) {
            case WEST:
                position.x = -label.getSize().x - labelSpacing;
                position.y = port.getSize().y + labelSpacing;
                break;
            case EAST:
                position.x = port.getSize().x + labelSpacing;
                position.y = port.getSize().y + labelSpacing;
                break;
            case NORTH:
                position.x = port.getSize().x + labelSpacing;
                position.y = -label.getSize().y - labelSpacing;
                break;
            case SOUTH:
                position.x = port.getSize().x + labelSpacing;
                position.y = port.getSize().y + labelSpacing;
                break;
            }
        }
        label.setPosition(position);
    }

    /**
     * Calculates the port's margins such that its label is part of them and sets them accordingly.
     * 
     * <p><i>Note:</i> We currently only support one label per port.</p>
     * 
     * @param port the port whose margins to calculate.
     */
    private void calculateAndSetPortMargins(final PortAdapter<?> port) {
        // Get the port's label, if any
        Iterator<LabelAdapter<?>> labelIter = port.getLabels().iterator();
        if (labelIter.hasNext()) {
            Rectangle portBox = new Rectangle(
                    0.0,
                    0.0,
                    port.getSize().x,
                    port.getSize().y);
            
            // We only support one label, so retrieve it
            LabelAdapter<?> label = labelIter.next();
            Rectangle labelBox = new Rectangle(
                    label.getPosition().x,
                    label.getPosition().y,
                    label.getSize().x,
                    label.getSize().y);
            
            // Calculate the union of the two bounding boxes and calculate the margins
            portBox.union(labelBox);

            Margins margin = new Margins(port.getMargin());
            margin.top = -portBox.y;
            margin.bottom = portBox.y + portBox.height - port.getSize().y;
            margin.left = -portBox.x;
            margin.right = portBox.x + portBox.width - port.getSize().x;
            port.setMargin(margin);
        }
    }
    
    /**
     * Calculates the width of ports on the northern and southern sides, and the height of ports on the
     * western and eastern sides of the given node. The information are stored in the class fields and
     * are used later on when calculating the minimum node size and when placing ports.
     * 
     * @param node the node to calculate the port information for.
     * @param accountForLabels if {@code true}, the port labels will be taken into account
     *                         when calculating the port information.
     */
    private void calculatePortInformation(final NodeAdapter<?> node, final boolean accountForLabels) {
        // Iterate over the ports
        for (PortAdapter<?> port : node.getPorts()) {
            switch (port.getSide()) {
            case WEST:
                westPortsCount++;
                westPortsHeight += port.getSize().y
                    + (accountForLabels ? port.getMargin().bottom + port.getMargin().top : 0.0);
                break;
            case EAST:
                eastPortsCount++;
                eastPortsHeight += port.getSize().y
                    + (accountForLabels ? port.getMargin().bottom + port.getMargin().top : 0.0);
                break;
            case NORTH:
                northPortsCount++;
                northPortsWidth += port.getSize().x
                    + (accountForLabels ? port.getMargin().left + port.getMargin().right : 0.0);
                break;
            case SOUTH:
                southPortsCount++;
                southPortsWidth += port.getSize().x
                    + (accountForLabels ? port.getMargin().left + port.getMargin().right : 0.0);
                break;
            }
        }
    }

    
    ///////////////////////////////////////////////////////////////////////////////
    // INSETS CALCULATION
    
    /**
     * Calculates the space required to accommodate all port labels and sets
     * {@link #requiredPortLabelSpace}. Also counts the number of ports on
     * each side of the node.
     * 
     * <p><i>Note:</i> We currently only support one label per port.</p>
     * 
     * @param node the node whose insets to calculate and to set.
     */
    private void calculateRequiredPortLabelSpace(final NodeAdapter<?> node) {
        // Iterate over the ports and look at their margins
        for (PortAdapter<?> port : node.getPorts()) {
            switch (port.getSide()) {
            case WEST:
                requiredPortLabelSpace.left =
                    Math.max(requiredPortLabelSpace.left, port.getMargin().right);
                break;
            case EAST:
                requiredPortLabelSpace.right =
                    Math.max(requiredPortLabelSpace.right, port.getMargin().left);
                break;
            case NORTH:
                requiredPortLabelSpace.top =
                    Math.max(requiredPortLabelSpace.top, port.getMargin().bottom);
                break;
            case SOUTH:
                requiredPortLabelSpace.bottom =
                    Math.max(requiredPortLabelSpace.bottom, port.getMargin().top);
                break;
            }
        }
    }

    /**
     * Calculates the space required to accommodate the node labels (if any) and sets
     * {@link #requiredNodeLabelSpace} as well as {@link #nodeLabelsBoundingBox}. If the labels are
     * placed at the top or at the bottom, the top or bottom insets are set. If it is centered
     * vertically, the left or right insets are set if the labels are horizontally aligned leftwards
     * or rightwards. If they are centered in both directions, no insets are set. If they are placed
     * outside the node, no insets are set.
     * 
     * @param node
     *            the node in question.
     * @param labelSpacing
     *            spacing between labels and other objects.
     */
    private void calculateRequiredNodeLabelSpace(final NodeAdapter<?> node, final double labelSpacing) {
        // Check if there are any labels
        if (!node.getLabels().iterator().hasNext()) {
            return;
        }

        // Compute a bounding box for all labels stacked vertically
        nodeLabelsBoundingBox.x = 0;
        nodeLabelsBoundingBox.y = 0;

        for (LabelAdapter<?> label : node.getLabels()) {
            nodeLabelsBoundingBox.x = Math.max(nodeLabelsBoundingBox.x, label.getSize().x);
            nodeLabelsBoundingBox.y += label.getSize().y + labelSpacing;
        }
        nodeLabelsBoundingBox.y -= labelSpacing;

        // Retrieve label placement policy
        EnumSet<NodeLabelPlacement> nodeLabelPlacement =
                node.getProperty(LayoutOptions.NODE_LABEL_PLACEMENT);

        // This method only sets insets if the node label is to be placed on the inside
        if (nodeLabelPlacement.contains(NodeLabelPlacement.INSIDE)) {
            // The primary distinction criterion is the vertical placement
            if (nodeLabelPlacement.contains(NodeLabelPlacement.V_TOP)) {
                requiredNodeLabelSpace.top = nodeLabelsBoundingBox.y + labelSpacing;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_BOTTOM)) {
                requiredNodeLabelSpace.bottom = nodeLabelsBoundingBox.y + labelSpacing;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_CENTER)) {
                // Check whether the label will be placed left or right
                if (nodeLabelPlacement.contains(NodeLabelPlacement.H_LEFT)) {
                    requiredNodeLabelSpace.left = nodeLabelsBoundingBox.x + labelSpacing;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_RIGHT)) {
                    requiredNodeLabelSpace.right = nodeLabelsBoundingBox.x + labelSpacing;
                }
            }
        }
    }

    
    ///////////////////////////////////////////////////////////////////////////////
    // NODE RESIZING
    
    /**
     * Resizes the given node subject to the sizing constraints and options.
     * 
     * @param node the node to resize.
     * @param portSpacing the minimum amount of space to be left between ports if there positions are
     *                    not fixed.
     * @param labelSpacing the amount of space to leave between labels and other objects.
     */
    private void resizeNode(final NodeAdapter<?> node, final double portSpacing,
            final double labelSpacing) {
        
        KVector nodeSize = node.getSize();
        KVector originalNodeSize = new KVector(nodeSize);
        EnumSet<SizeConstraint> sizeConstraint = node.getProperty(LayoutOptions.SIZE_CONSTRAINT);
        EnumSet<SizeOptions> sizeOptions = node.getProperty(LayoutOptions.SIZE_OPTIONS);
        PortConstraints portConstraints = node.getProperty(LayoutOptions.PORT_CONSTRAINTS);
        boolean accountForLabels = sizeConstraint.contains(SizeConstraint.PORT_LABELS);
        
        // If the size constraint is empty, we can't do anything
        if (sizeConstraint.isEmpty()) {
            return;
        }
        
        // It's not empty, so we will change the node size; we start by resetting the size to zero
        nodeSize.x = 0.0;
        nodeSize.y = 0.0;
        
        // Find out how large the node will have to be to accommodate all ports. If port
        // constraints are set to FIXED_RATIO, we can't do anything smart, really; in this
        // case we will just assume the original node size to be the minumum for ports
        KVector minSizeForPorts = null;
        switch (portConstraints) {
        case FREE:
        case FIXED_SIDE:
        case FIXED_ORDER:
            // Calculate the space necessary to accomodate all ports
            minSizeForPorts = calculatePortSpaceRequirements(node, portSpacing, accountForLabels);
            break;
        
        case FIXED_RATIO:
            // Keep original node size
            minSizeForPorts = new KVector(originalNodeSize);
            break;
        
        case FIXED_POS:
            // Find the maximum position of ports
            minSizeForPorts = calculateMaxPortPositions(node, accountForLabels);
            break;
        }
        
        // If the node size should take port space requirements into account, adjust it accordingly
        if (sizeConstraint.contains(SizeConstraint.PORTS)) {
            // Check if we have a minimum size required for all ports
            if (minSizeForPorts != null) {
                nodeSize.x = Math.max(nodeSize.x, minSizeForPorts.x);
                nodeSize.y = Math.max(nodeSize.y, minSizeForPorts.y);
            }
            
            // If we account for labels, we need to have the size account for labels placed on the
            // inside of the node (this only affects port label placement INSIDE; OUTSIDE is already
            // part of minSizeForPorts)
            if (accountForLabels) {
                nodeSize.x = Math.max(nodeSize.x,
                        requiredPortLabelSpace.left + requiredPortLabelSpace.right + portSpacing);
                nodeSize.y = Math.max(nodeSize.y,
                        requiredPortLabelSpace.top + requiredPortLabelSpace.bottom + portSpacing);
            }
        }
        
        // If the node label is to be accounted for, add its required space to the node size
        if (sizeConstraint.contains(SizeConstraint.NODE_LABELS)
                && node.getLabels().iterator().hasNext()) {
            
            EnumSet<NodeLabelPlacement> nodeLabelPlacement =
                    node.getProperty(LayoutOptions.NODE_LABEL_PLACEMENT);
            
            // Check if the label is to be placed inside or outside the node
            if (nodeLabelPlacement.contains(NodeLabelPlacement.INSIDE)) {
                nodeSize.x += requiredNodeLabelSpace.left + requiredNodeLabelSpace.right;
                nodeSize.y += requiredNodeLabelSpace.top + requiredNodeLabelSpace.bottom;

                // For center placement, the insets don't cover everything
                if (nodeLabelPlacement.contains(NodeLabelPlacement.V_CENTER)) {
                    nodeSize.y += nodeLabelsBoundingBox.y + 2 * labelSpacing;
                }

                if (nodeLabelPlacement.contains(NodeLabelPlacement.H_CENTER)) {
                    nodeSize.x += nodeLabelsBoundingBox.x + 2 * labelSpacing;
                }
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.OUTSIDE)) {
                // The node must be at least as high or wide as the label
                if (nodeLabelPlacement.contains(NodeLabelPlacement.V_TOP)
                        || nodeLabelPlacement.contains(NodeLabelPlacement.V_BOTTOM)) {
                    
                    nodeSize.x = Math.max(nodeSize.x, nodeLabelsBoundingBox.x);
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_CENTER)) {
                    nodeSize.y = Math.max(nodeSize.y, nodeLabelsBoundingBox.y);
                }
            }
        }
        
        // Respect minimum size
        if (sizeConstraint.contains(SizeConstraint.MINIMUM_SIZE)) {
            double minWidth = node.getProperty(LayoutOptions.MIN_WIDTH);
            double minHeight = node.getProperty(LayoutOptions.MIN_HEIGHT);
            
            // If we are to use default minima, check if the values are properly set
            if (sizeOptions.contains(SizeOptions.DEFAULT_MINIMUM_SIZE)) {
                if (minWidth <= 0) {
                    minWidth = KimlUtil.DEFAULT_MIN_WIDTH;
                }
                
                if (minHeight <= 0) {
                    minHeight = KimlUtil.DEFAULT_MIN_HEIGHT;
                }
            }
            
            // We might have to take the insets into account
            if (sizeOptions.contains(SizeOptions.MINIMUM_SIZE_ACCOUNTS_FOR_INSETS)) {
                if (minWidth > 0) {
                    nodeSize.x = Math.max(nodeSize.x,
                            minWidth
                            + requiredPortLabelSpace.left
                            + requiredPortLabelSpace.right);
                }
                
                if (minHeight > 0) {
                    nodeSize.y = Math.max(nodeSize.y,
                            minHeight
                            + requiredPortLabelSpace.top
                            + requiredPortLabelSpace.bottom);
                }
            } else {
                if (minWidth > 0) {
                    nodeSize.x = Math.max(nodeSize.x, minWidth);
                }
                
                if (minHeight > 0) {
                    nodeSize.y = Math.max(nodeSize.y, minHeight);
                }
            }
        }
        // apply the calculated node size back to the wrapped node
        node.setSize(nodeSize);
    }
    
    /**
     * Calculate how much space the ports will require at a minimum if they can be freely distributed on
     * their respective node side. The x coordinate of the returned vector will be the minimum width
     * required by the ports on the northern and southern side. The y coordinate, in turn, will be the
     * minimum height required by the ports on the western and eastern side. This may include the space
     * required for port labels placed outside of the node. If port labels are placed inside, their
     * space requirements are not included in the result.
     * 
     * @param node the node to calculate the minimum size for.
     * @param portSpacing the minimum amount of space to be left between ports if there positions are
     *                    not fixed.
     * @param accountForLabels if {@code true}, the port labels will be taken into account
     *                         when calculating the space requirements.
     * @return minimum size.
     */
    private KVector calculatePortSpaceRequirements(final NodeAdapter<?> node, final double portSpacing,
            final boolean accountForLabels) {
        
        // Calculate the additional port space to be left around the set of ports on each side. If this
        // is not set, we assume the spacing to be the minimum space left between ports
        double additionalWidth;
        double additionalHeight;
        
        Margins additionalPortSpace = node.getProperty(LayoutOptions.ADDITIONAL_PORT_SPACE);
        if (additionalPortSpace == null) {
            additionalWidth = 2 * portSpacing;
            additionalHeight = 2 * portSpacing;
        } else {
            additionalWidth = additionalPortSpace.left + additionalPortSpace.right;
            additionalHeight = additionalPortSpace.top + additionalPortSpace.bottom;
        }
        
        // Calculate the required width and height, taking the necessary spacing between (and around)
        // the ports into consideration as well
        double requiredWidth = Math.max(
                northPortsCount > 0
                    ? additionalWidth + (northPortsCount - 1) * portSpacing + northPortsWidth
                    : 0.0,
                southPortsCount > 0
                    ? additionalWidth + (southPortsCount - 1) * portSpacing + southPortsWidth
                    : 0.0);
        double requiredHeight = Math.max(
                westPortsCount > 0
                    ? additionalHeight + (westPortsCount - 1) * portSpacing + westPortsHeight
                    : 0.0,
                eastPortsCount > 0
                    ? additionalHeight + (eastPortsCount - 1) * portSpacing + eastPortsHeight
                    : 0.0);
        
        return new KVector(requiredWidth, requiredHeight);
    }
    
    /**
     * For fixed node positions, returns the minimum size of the node to contain all ports.
     * 
     * @param node the node to calculate the minimum size for.
     * @param accountForLabels
     * @return
     */
    private KVector calculateMaxPortPositions(final NodeAdapter<?> node, 
            final boolean accountForLabels) {
        
        // Port positions must be fixed for this method to be called
        assert node.getProperty(LayoutOptions.PORT_CONSTRAINTS) == PortConstraints.FIXED_POS;
        
        KVector result = new KVector();
        
        // Iterate over the ports
        for (PortAdapter<?> port : node.getPorts()) {
            switch (port.getSide()) {
            case WEST:
            case EAST:
                result.y = Math.max(result.y,
                        port.getPosition().y
                        + port.getSize().y
                        + (accountForLabels ? port.getMargin().bottom : 0.0));
                break;
                
            case NORTH:
            case SOUTH:
                result.x = Math.max(result.x,
                        port.getPosition().x
                        + port.getSize().x
                        + (accountForLabels ? port.getMargin().right : 0.0));
                break;
            }
        }
        
        return result;
    }

    
    ///////////////////////////////////////////////////////////////////////////////
    // PORT PLACEMENT

    /**
     * Places the given node's ports. If the node wasn't resized at all and port constraints are set
     * to either {@link PortConstraints#FIXED_RATIO} or {@link PortConstraints#FIXED_POS}, the port
     * positions are not touched.
     * 
     * @param node the node whose ports to place.
     * @param originalNodeSize the node's size before it was (possibly) resized.
     */
    private void placePorts(final NodeAdapter<?> node, final KVector originalNodeSize) {
        PortConstraints portConstraints = node.getProperty(LayoutOptions.PORT_CONSTRAINTS);
        
        if (portConstraints == PortConstraints.FIXED_POS) {
            // Fixed Position
            placeFixedPosNodePorts(node);
        } else if (portConstraints == PortConstraints.FIXED_RATIO) {
            // Fixed Ratio
            placeFixedRatioNodePorts(node);
        } else {
            // Free, Fixed Side, Fixed Order
            if (node.getProperty(LayoutOptions.HYPERNODE)
                    || (node.getSize().x == 0 && node.getSize().y == 0)) {
                
                placeHypernodePorts(node);
            } else {
                placeNodePorts(node);
            }
        }
    }
    
    /**
     * Places the ports of a node assuming that the port constraints are set to fixed port positions.
     * Ports still need to be placed, though, because the node may have been resized.
     * 
     * @param node the node whose ports to place.
     */
    private void placeFixedPosNodePorts(final NodeAdapter<?> node) {
        KVector nodeSize = node.getSize();

        for (PortAdapter<?> port : node.getPorts()) {
            Float portOffset = port.getProperty(LayoutOptions.OFFSET);
            if (portOffset == null) {
                portOffset = 0f;
            }
            
            KVector position = new KVector(port.getPosition());
            switch (port.getSide()) {
            case WEST:
                position.x = -port.getSize().x - portOffset;
                break;
            case EAST:
                position.x = nodeSize.x + portOffset;
                break;
            case NORTH:
                position.y = -port.getSize().y - portOffset;
                break;
            case SOUTH:
                position.y = nodeSize.y + portOffset;
                break;
            }
            port.setPosition(position);
        }
    }

    /**
     * Places the ports of a node keeping the ratio between their position and the length of their
     * respective side intact.
     * 
     * @param node the node whose ports to place.
     */
    private void placeFixedRatioNodePorts(final NodeAdapter<?> node) {
        KVector nodeSize = node.getSize();

        // Adjust port positions depending on port side. Eastern ports have to have their x coordinate
        // set to the node's current width; the same goes for the y coordinate of southern ports
        for (PortAdapter<?> port : node.getPorts()) {
            Float portOffset = port.getProperty(LayoutOptions.OFFSET);
            if (portOffset == null) {
                portOffset = 0f;
            }
            
            switch (port.getSide()) {
            case WEST:
                port.getPosition().y =
                        nodeSize.y * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().x = -port.getSize().x - portOffset;
                break;
            case EAST:
                port.getPosition().y =
                        nodeSize.y * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().x = nodeSize.x + portOffset;
                break;
            case NORTH:
                port.getPosition().x =
                        nodeSize.x * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().y = -port.getSize().y - portOffset;
                break;
            case SOUTH:
                port.getPosition().x =
                        nodeSize.x * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().y = nodeSize.y + portOffset;
                break;
            }
            switch (port.getSide()) {
            case WEST:
                port.getPosition().y =
                        nodeSize.y * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().x = -port.getSize().x - portOffset;
                break;
            case EAST:
                port.getPosition().y =
                        nodeSize.y * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().x = nodeSize.x + portOffset;
                break;
            case NORTH:
                port.getPosition().x =
                        nodeSize.x * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().y = -port.getSize().y - portOffset;
                break;
            case SOUTH:
                port.getPosition().x =
                        nodeSize.x * port.getProperty(PORT_RATIO_OR_POSITION);
                port.getPosition().y = nodeSize.y + portOffset;
                break;
            }
        }
    }
    
    /**
     * Places the ports of a node, assuming that the ports are not fixed in their position or ratio.
     * 
     * @param node the node whose ports to place.
     */
    private void placeNodePorts(final NodeAdapter<?> node) {
        KVector nodeSize = node.getSize();
        boolean accountForLabels =
                node.getProperty(LayoutOptions.SIZE_CONSTRAINT).contains(SizeConstraint.PORT_LABELS);
        
        // Let someone compute the port placement data we'll need
        PortPlacementData placementData = computePortPlacementData(node);
        
        // Arrange the ports
        for (PortAdapter<?> port : node.getPorts()) {
            Float portOffset = port.getProperty(LayoutOptions.OFFSET);
            if (portOffset == null) {
                portOffset = 0f;
            }
            KVector portSize = port.getSize();
            Margins portMargins = port.getMargin();
            
            KVector position = new KVector(port.getPosition());
            switch (port.getSide()) {
            case WEST:
                position.x = -portSize.x - portOffset;
                position.y = placementData.westY - portSize.y
                        - (accountForLabels ? portMargins.bottom : 0.0);
                placementData.westY -= placementData.westGapSize + portSize.y
                        + (accountForLabels ? portMargins.top + portMargins.bottom : 0.0);
                break;
            case EAST:
                position.x = nodeSize.x + portOffset;
                position.y = placementData.eastY
                        + (accountForLabels ? portMargins.top : 0.0);
                placementData.eastY += placementData.eastGapSize + portSize.y
                        + (accountForLabels ? portMargins.top + portMargins.bottom : 0.0);
                break;
            case NORTH:
                position.x = placementData.northX
                        + (accountForLabels ? portMargins.left : 0.0);
                position.y = -port.getSize().y - portOffset;
                placementData.northX += placementData.northGapSize + portSize.x
                        + (accountForLabels ? portMargins.left + portMargins.right : 0.0);
                break;
            case SOUTH:
                position.x = placementData.southX - portSize.x
                        - (accountForLabels ? portMargins.right : 0.0);
                position.y = nodeSize.y + portOffset;
                placementData.southX -= placementData.southGapSize + portSize.x
                        + (accountForLabels ? portMargins.left + portMargins.right : 0.0);
                break;
            }
            port.setPosition(position);
        }
    }
    
    /**
     * Computes the port placement data for the given node.
     * 
     * @param node the node to compute the placement data for.
     * @return the port placement data.
     */
    private PortPlacementData computePortPlacementData(final NodeAdapter<?> node) {
        PortPlacementData data = new PortPlacementData();
        KVector nodeSize = node.getSize();
        
        // The way we calculate everything depends on whether any additional port space is specified
        Margins additionalPortSpace = node.getProperty(LayoutOptions.ADDITIONAL_PORT_SPACE);
        if (additionalPortSpace == null) {
            // No additional port spacing, so we simply distribute the ports on each side
            data.westGaps = westPortsCount + 1;
            data.eastGaps = eastPortsCount + 1;
            data.northGaps = northPortsCount + 1;
            data.southGaps = southPortsCount + 1;

            // Compute the space to be left between the ports
            // Note: If the size constraints of this node are empty, the height and width of the ports
            // on each side are zero. That is intentional: if this wasn't the case, bad things would
            // happen if the ports would actually need more size than the node at its current (unchanged)
            // size would be able to provide.
            data.westGapSize = (nodeSize.y - westPortsHeight) / data.westGaps;
            data.eastGapSize = (nodeSize.y - eastPortsHeight) / data.eastGaps;
            data.northGapSize = (nodeSize.x - northPortsWidth) / data.northGaps;
            data.southGapSize = (nodeSize.x - southPortsWidth) / data.southGaps;
            
            // Compute the coordinate of the first port on each side
            data.westY = nodeSize.y - data.westGapSize;
            data.eastY = data.eastGapSize;
            data.northX = data.northGapSize;
            data.southX = nodeSize.x - data.southGapSize;
        } else {
            // Calculate how much space on each side may actually be used by ports
            double usableHeight = nodeSize.y - additionalPortSpace.top - additionalPortSpace.bottom;
            double usableWidth = nodeSize.x - additionalPortSpace.left - additionalPortSpace.right;
            
            // Calculate how many gaps we have between ports (this is usually one less than the number of
            // ports we have, but if it's just a single port, we have two gaps that surround it)
            data.westGaps = westPortsCount == 1 ? 2 : westPortsCount - 1;
            data.eastGaps = eastPortsCount == 1 ? 2 : eastPortsCount - 1;
            data.northGaps = northPortsCount == 1 ? 2 : northPortsCount - 1;
            data.southGaps = southPortsCount == 1 ? 2 : southPortsCount - 1;

            // Compute the space to be left between the ports
            // Note: If the size constraints of this node are empty, the height and width of the ports
            // on each side are zero. That is intentional: if this wasn't the case, bad things would
            // happen if the ports would actually need more size than the node at its current (unchanged)
            // size would be able to provide.
            data.westGapSize = (usableHeight - westPortsHeight) / data.westGaps;
            data.eastGapSize = (usableHeight - eastPortsHeight) / data.eastGaps;
            data.northGapSize = (usableWidth - northPortsWidth) / data.northGaps;
            data.southGapSize = (usableWidth - southPortsWidth) / data.southGaps;
            
            // Compute the coordinate of the first port on each side
            data.westY = nodeSize.y
                    - additionalPortSpace.bottom
                    - (westPortsCount == 1 ? data.westGapSize : 0);
            data.eastY = additionalPortSpace.top
                    + (eastPortsCount == 1 ? data.eastGapSize : 0);
            data.northX = additionalPortSpace.left
                    + (northPortsCount == 1 ? data.northGapSize : 0);
            data.southX = nodeSize.x
                    - additionalPortSpace.right
                    - (southPortsCount == 1 ? data.southGapSize : 0);
        }
        
        return data;
    }
    
    /**
     * Places the ports of a hypernode.
     * 
     * @param node the hypernode whose ports to place.
     */
    private void placeHypernodePorts(final NodeAdapter<?> node) {
        for (PortAdapter<?> port : node.getPorts()) {
            KVector position = new KVector(port.getPosition());
            switch (port.getSide()) {
            case WEST:
                position.x = 0;
                position.y = node.getSize().y / 2;
                break;
            case EAST:
                position.x = node.getSize().x;
                position.y = node.getSize().y / 2;
                break;
            case NORTH:
                position.x = node.getSize().x / 2;
                position.y = 0;
                break;
            case SOUTH:
                position.x = node.getSize().x / 2;
                position.y = node.getSize().y;
                break;
            }
            port.setPosition(position);
        }
    }

    
    ///////////////////////////////////////////////////////////////////////////////
    // PLACING NODE LABELS
    
    /**
     * Calculates the position of the node's labels.
     * 
     * @param node the node whose labels to place.
     * @param labelSpacing spacing between labels and other objects.
     */
    private void placeNodeLabels(final NodeAdapter<?> node, final double labelSpacing) {
        // Retrieve the first node label, if any
        if (!node.getLabels().iterator().hasNext()) {
            return;
        }
        
        // Top left position where the node labels will start to be placed in a vertical stack
        KVector labelGroupPos = new KVector();
        
        // Retrieve label placement policy
        EnumSet<NodeLabelPlacement> nodeLabelPlacement =
                node.getProperty(LayoutOptions.NODE_LABEL_PLACEMENT);
        
        // This method only sets insets if the node label is to be placed on the inside
        if (nodeLabelPlacement.contains(NodeLabelPlacement.INSIDE)) {
            // Y coordinate
            if (nodeLabelPlacement.contains(NodeLabelPlacement.V_TOP)) {
                labelGroupPos.y = requiredPortLabelSpace.top + labelSpacing;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_CENTER)) {
                labelGroupPos.y = (node.getSize().y - nodeLabelsBoundingBox.y) / 2.0;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_BOTTOM)) {
                labelGroupPos.y = node.getSize().y - requiredPortLabelSpace.bottom
                        - nodeLabelsBoundingBox.y - labelSpacing;
            }
            
            // X coordinate
            if (nodeLabelPlacement.contains(NodeLabelPlacement.H_LEFT)) {
                labelGroupPos.x = requiredPortLabelSpace.left + labelSpacing;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_CENTER)) {
                labelGroupPos.x = (node.getSize().x - nodeLabelsBoundingBox.x) / 2.0;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_RIGHT)) {
                labelGroupPos.x = node.getSize().x - requiredPortLabelSpace.right
                        - nodeLabelsBoundingBox.x - labelSpacing;
            }
        } else if (nodeLabelPlacement.contains(NodeLabelPlacement.OUTSIDE)) {
            // TODO: Outside placement doesn't take ports and port labels into account yet.
            
            // Different placement logic depending on whether horizontal or vertical placement
            // is prioritized
            if (nodeLabelPlacement.contains(NodeLabelPlacement.H_PRIORITY)) {
                boolean leftOrRight = false;
                
                // X coordinate
                if (nodeLabelPlacement.contains(NodeLabelPlacement.H_LEFT)) {
                    labelGroupPos.x = -(nodeLabelsBoundingBox.x + labelSpacing);
                    leftOrRight = true;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_CENTER)) {
                    labelGroupPos.x = (node.getSize().x - nodeLabelsBoundingBox.x) / 2.0;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_RIGHT)) {
                    labelGroupPos.x = node.getSize().x + labelSpacing;
                    leftOrRight = true;
                }
                
                // Y coordinate
                if (nodeLabelPlacement.contains(NodeLabelPlacement.V_TOP)) {
                    if (leftOrRight) {
                        labelGroupPos.y = 0;
                    } else {
                        labelGroupPos.y = -(nodeLabelsBoundingBox.y + labelSpacing);
                    }
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_CENTER)) {
                    labelGroupPos.y = (node.getSize().y - nodeLabelsBoundingBox.y) / 2.0;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_BOTTOM)) {
                    if (leftOrRight) {
                        labelGroupPos.y = node.getSize().y - nodeLabelsBoundingBox.y;
                    } else {
                        labelGroupPos.y = node.getSize().y + labelSpacing;
                    }
                }
            } else {
                boolean topOrBottom = false;
                
                // Y coordinate
                if (nodeLabelPlacement.contains(NodeLabelPlacement.V_TOP)) {
                    labelGroupPos.y = -(nodeLabelsBoundingBox.y + labelSpacing);
                    topOrBottom = true;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_CENTER)) {
                    labelGroupPos.y = (node.getSize().y - nodeLabelsBoundingBox.y) / 2.0;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.V_BOTTOM)) {
                    labelGroupPos.y = node.getSize().y + labelSpacing;
                    topOrBottom = true;
                }
                
                // X coordinate
                if (nodeLabelPlacement.contains(NodeLabelPlacement.H_LEFT)) {
                    if (topOrBottom) {
                        labelGroupPos.x = 0;
                    } else {
                        labelGroupPos.x = -(nodeLabelsBoundingBox.x + labelSpacing);
                    }
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_CENTER)) {
                    labelGroupPos.x = (node.getSize().x - nodeLabelsBoundingBox.x) / 2.0;
                } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_RIGHT)) {
                    if (topOrBottom) {
                        labelGroupPos.x = node.getSize().x - nodeLabelsBoundingBox.x;
                    } else {
                        labelGroupPos.x = node.getSize().x + labelSpacing;
                    }
                }
            }
        }
        
        // Place labels
        applyNodeLabelPositions(node, labelGroupPos, labelSpacing);
    }
    
    /**
     * Places the given node's labels in a vertical stack, starting at the given position.
     * 
     * @param node the node whose labels are to be placed.
     * @param startPosition coordinates where the first label is to be placed.
     * @param labelSpacing space to be left between the labels.
     */
    private void applyNodeLabelPositions(final NodeAdapter<?> node, final KVector startPosition,
            final double labelSpacing) {
        
        // The horizontal alignment depends on where the labels are placed exactly
        EnumSet<NodeLabelPlacement> nodeLabelPlacement =
                node.getProperty(LayoutOptions.NODE_LABEL_PLACEMENT);
        NodeLabelPlacement horizontalPlacement = NodeLabelPlacement.H_CENTER;
        
        if (nodeLabelPlacement.contains(NodeLabelPlacement.INSIDE)) {
            // Inside placement
            if (nodeLabelPlacement.contains(NodeLabelPlacement.H_LEFT)) {
                horizontalPlacement = NodeLabelPlacement.H_LEFT;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_RIGHT)) {
                horizontalPlacement = NodeLabelPlacement.H_RIGHT;
            }
        } else {
            // Outside placement; alignment is reversed
            if (nodeLabelPlacement.contains(NodeLabelPlacement.H_LEFT)) {
                horizontalPlacement = NodeLabelPlacement.H_RIGHT;
            } else if (nodeLabelPlacement.contains(NodeLabelPlacement.H_RIGHT)) {
                horizontalPlacement = NodeLabelPlacement.H_LEFT;
            }
        }
        
        // We have to keep track of the current y coordinate
        double currentY = startPosition.y;
        
        // Place all labels
        for (LabelAdapter<?> label : node.getLabels()) {
            KVector position = new KVector(label.getPosition());
            // Apply y coordinate
            //label.getPosition().y = currentY;
            position.y = currentY;
            
            // The x coordinate depends on the H_xxx constants
            if (horizontalPlacement == NodeLabelPlacement.H_LEFT) {
                position.x = startPosition.x;
            } else if (horizontalPlacement == NodeLabelPlacement.H_CENTER) {
                position.x = startPosition.x
                        + (nodeLabelsBoundingBox.x - label.getSize().x) / 2.0; 
            } else if (horizontalPlacement == NodeLabelPlacement.H_RIGHT) {
                position.x = startPosition.x + nodeLabelsBoundingBox.x - label.getSize().x;
            }
            
            label.setPosition(position);
            
            // Update y position
            currentY += label.getSize().y + labelSpacing;
        }
    }

    
    ///////////////////////////////////////////////////////////////////////////////
    // PORT PLACEMENT UTILITIES
    
    /**
     * Holds information necessary to place the ports on each side. Since a lot of information are
     * necessary, we define a small data holder class just for them. Not all of the fields specified
     * here are always required.
     * 
     * @author cds
     */
    private static final class PortPlacementData {
        // The number of gaps between the ports (this is usually one less than the number of ports we
        // have, but if it's just a single port, we have two gaps that surround it)
        private double westGaps;
        private double eastGaps;
        private double northGaps;
        private double southGaps;
        
        // The size of each gap on the different sides
        private double westGapSize;
        private double eastGapSize;
        private double northGapSize;
        private double southGapSize;
        
        // The position of the next port on each side
        private double westY;
        private double eastY;
        private double northX;
        private double southX;
    }
}
