/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.klighd.xtext.transformations;

import com.google.common.base.Objects;
import com.google.common.base.Strings;
import com.google.common.collect.ImmutableList;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KGraphElement;
import de.cau.cs.kieler.core.kgraph.KLabel;
import de.cau.cs.kieler.core.kgraph.KLabeledGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.krendering.KColor;
import de.cau.cs.kieler.core.krendering.KEllipse;
import de.cau.cs.kieler.core.krendering.KPolyline;
import de.cau.cs.kieler.core.krendering.KRectangle;
import de.cau.cs.kieler.core.krendering.KRendering;
import de.cau.cs.kieler.core.krendering.KRenderingFactory;
import de.cau.cs.kieler.core.krendering.KRenderingLibrary;
import de.cau.cs.kieler.core.krendering.KRenderingRef;
import de.cau.cs.kieler.core.krendering.KRoundedRectangle;
import de.cau.cs.kieler.core.krendering.KSpline;
import de.cau.cs.kieler.core.krendering.KStyleHolder;
import de.cau.cs.kieler.core.krendering.KText;
import de.cau.cs.kieler.core.krendering.extensions.KColorExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KLibraryExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KPolylineExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.core.properties.IProperty;
import de.cau.cs.kieler.core.properties.Property;
import de.cau.cs.kieler.core.util.Pair;
import de.cau.cs.kieler.kiml.klayoutdata.KIdentifier;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeLabelPlacement;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.NodeLabelPlacement;
import de.cau.cs.kieler.kiml.options.SizeConstraint;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.klighd.SynthesisOption;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Synthesizes a copy of the given {@code KNode} and adds default stuff.
 * 
 * <p>Default stuff, in this case, is the following:</p>
 * <ul>
 *   <li>If a node has no label, but an ID, a label with the ID as its text is added.</li>
 *   <li>If a node label is present, node label placement defaults to inner centered placement.</li>
 *   <li>Edges without attached {@code KRendering}s are provided with a {@code KPolyline} with
 *       an arrowhead at the end. (the default would be a simple polyline without arrowhead)</li>
 *   <li>Labels without attached {@code KRendering}s are provided with a default {@code KText}
 *       rendering that ensures that label sizes will be calculated correctly.</li>
 *   <li>Edge labels without edge label placement get a shiny new {@link EdgeLabelPlacement#CENTER}
 *       placement.</li>
 * </ul>
 * 
 * @author cds
 * @author uru
 */
@SuppressWarnings("all")
public class KGraphDiagramSynthesis extends AbstractDiagramSynthesis<KNode> {
  /**
   * ID this synthesis is registered with KLighD with.
   */
  public static String TRANSFORMATION_ID = "de.cau.cs.kieler.klighd.xtext.transformations.KGraphDiagramSynthesis";
  
  /**
   * Whether the model wants default defaults. This property is no layout option, not registered with
   * KIML, and has to be loaded explicitly by the
   * {@link de.cau.cs.kieler.core.kgraph.text.KGraphResource KGraphResource}.
   */
  private final static IProperty<Boolean> DEFAULTS_PROPERTY = new Property<Boolean>("de.cau.cs.kieler.kgraphsynthesis.defaults", Boolean.valueOf(false));
  
  private final static String DEFAULTS_AS_IN_MODEL = "As in Model";
  
  private final static String DEFAULTS_ON = "On";
  
  private final static String DEFAULTS_OFF = "Off";
  
  /**
   * Synthesis option specifying whether default values should be used. Default values are, eg, node
   * size if not specified and port ids as labels if no labels exist.
   */
  private final static SynthesisOption DEFAULTS = SynthesisOption.createChoiceOption(
    "Default Values", 
    ImmutableList.<String>of(KGraphDiagramSynthesis.DEFAULTS_AS_IN_MODEL, KGraphDiagramSynthesis.DEFAULTS_ON, KGraphDiagramSynthesis.DEFAULTS_OFF), KGraphDiagramSynthesis.DEFAULTS_AS_IN_MODEL);
  
  private final static SynthesisOption STYLE = SynthesisOption.createChoiceOption(
    "Style", 
    ImmutableList.<String>of("Boring", "Stylish", "Hello Kitty"), "Boring");
  
  @Inject
  @Extension
  private KPolylineExtensions _kPolylineExtensions;
  
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  @Inject
  @Extension
  private KColorExtensions _kColorExtensions;
  
  @Inject
  @Extension
  private KLibraryExtensions _kLibraryExtensions;
  
  /**
   * Rendering factory used to create KRendering model instances.
   */
  private static KRenderingFactory renderingFactory = KRenderingFactory.eINSTANCE;
  
  /**
   * The value for the defaults property as specified in the model.
   */
  private boolean defaults = false;
  
  /**
   * Default rendering for nodes.
   */
  private KRendering defaultNodeRendering;
  
  /**
   * Default rendering for polyline edges.
   */
  private KRendering defaultPolylineRendering;
  
  /**
   * Default rendering for spline edges.
   */
  private KRendering defaultSplineRendering;
  
  /**
   * {@inheritDoc}
   */
  public List<Pair<IProperty<?>, List<?>>> getDisplayedLayoutOptions() {
    return ImmutableList.<Pair<IProperty<?>, List<?>>>of();
  }
  
  /**
   * {@inheritDoc}
   */
  public List<SynthesisOption> getDisplayedSynthesisOptions() {
    return ImmutableList.<SynthesisOption>of(
      KGraphDiagramSynthesis.DEFAULTS, KGraphDiagramSynthesis.STYLE);
  }
  
  /**
   * Transforms the given graph into an equivalent graph that may be enriched with additional
   * rendering information.
   * 
   * @param graph the graph to transform.
   * @return the possibly enriched graph.
   */
  public KNode transform(final KNode graph) {
    final EcoreUtil.Copier copier = new EcoreUtil.Copier();
    EObject _copy = copier.copy(graph);
    final KNode result = ((KNode) _copy);
    copier.copyReferences();
    try {
      KLayoutData _data = graph.<KLayoutData>getData(KLayoutData.class);
      Boolean _property = _data.<Boolean>getProperty(KGraphDiagramSynthesis.DEFAULTS_PROPERTY);
      this.defaults = (_property).booleanValue();
    } catch (final Throwable _t) {
      if (_t instanceof ClassCastException) {
        final ClassCastException cce = (ClassCastException)_t;
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
    KRenderingLibrary library = result.<KRenderingLibrary>getData(KRenderingLibrary.class);
    boolean _equals = Objects.equal(library, null);
    if (_equals) {
      KRenderingLibrary _createKRenderingLibrary = KGraphDiagramSynthesis.renderingFactory.createKRenderingLibrary();
      library = _createKRenderingLibrary;
      EList<KGraphData> _data_1 = result.getData();
      _data_1.add(library);
    }
    Object _objectValue = this.getObjectValue(KGraphDiagramSynthesis.STYLE);
    boolean _matched = false;
    if (!_matched) {
      if (Objects.equal(_objectValue, "Stylish")) {
        _matched=true;
        this.initStylishFactory(library);
      }
    }
    if (!_matched) {
      if (Objects.equal(_objectValue, "Hello Kitty")) {
        _matched=true;
        this.initHelloKittyFactory(library);
      }
    }
    if (!_matched) {
      this.initBoringFactory(library);
    }
    Set<Map.Entry<EObject, EObject>> _entrySet = copier.entrySet();
    for (final Map.Entry<EObject, EObject> entry : _entrySet) {
      EObject _value = entry.getValue();
      EObject _key = entry.getKey();
      this.<EObject>associateWith(_value, _key);
    }
    this.recursivelyEnrichRendering(result);
    return result;
  }
  
  private boolean initEdgeRenderings(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      KPolyline _createKPolyline = KGraphDiagramSynthesis.renderingFactory.createKPolyline();
      final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
        public void apply(final KPolyline it) {
          it.setId("DefaultEdgeRendering");
          KGraphDiagramSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
          KGraphDiagramSynthesis.this._kPolylineExtensions.addJunctionPointDecorator(it);
        }
      };
      KPolyline _doubleArrow = ObjectExtensions.<KPolyline>operator_doubleArrow(_createKPolyline, _function);
      this.defaultPolylineRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _renderings.add(this.defaultPolylineRendering);
      KSpline _createKSpline = KGraphDiagramSynthesis.renderingFactory.createKSpline();
      final Procedure1<KSpline> _function_1 = new Procedure1<KSpline>() {
        public void apply(final KSpline it) {
          it.setId("SplineEdgeRendering");
          KGraphDiagramSynthesis.this._kPolylineExtensions.addHeadArrowDecorator(it);
        }
      };
      KSpline _doubleArrow_1 = ObjectExtensions.<KSpline>operator_doubleArrow(_createKSpline, _function_1);
      this.defaultSplineRendering = _doubleArrow_1;
      EList<KStyleHolder> _renderings_1 = library.getRenderings();
      _xblockexpression = _renderings_1.add(this.defaultSplineRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initBoringFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KRectangle _createKRectangle = KGraphDiagramSynthesis.renderingFactory.createKRectangle();
      final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
        public void apply(final KRectangle it) {
          it.setId("DefaultNodeRendering");
        }
      };
      KRectangle _doubleArrow = ObjectExtensions.<KRectangle>operator_doubleArrow(_createKRectangle, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initStylishFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KRoundedRectangle _createKRoundedRectangle = KGraphDiagramSynthesis.renderingFactory.createKRoundedRectangle();
      final Procedure1<KRoundedRectangle> _function = new Procedure1<KRoundedRectangle>() {
        public void apply(final KRoundedRectangle it) {
          it.setId("DefaultNodeRendering");
          it.setCornerWidth(5);
          it.setCornerHeight(5);
          KColor _color = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#EBF6FA");
          KColor _color_1 = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#C4E3F3");
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KRoundedRectangle>setBackgroundGradient(it, _color, _color_1, 90);
        }
      };
      KRoundedRectangle _doubleArrow = ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createKRoundedRectangle, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  private boolean initHelloKittyFactory(final KRenderingLibrary library) {
    boolean _xblockexpression = false;
    {
      this.initEdgeRenderings(library);
      KEllipse _createKEllipse = KGraphDiagramSynthesis.renderingFactory.createKEllipse();
      final Procedure1<KEllipse> _function = new Procedure1<KEllipse>() {
        public void apply(final KEllipse it) {
          it.setId("DefaultNodeRendering");
          KColor _color = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#FFEEEE");
          KColor _color_1 = KGraphDiagramSynthesis.this._kColorExtensions.getColor("#FFBBBB");
          KGraphDiagramSynthesis.this._kRenderingExtensions.<KEllipse>setBackgroundGradient(it, _color, _color_1, 90);
        }
      };
      KEllipse _doubleArrow = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse, _function);
      this.defaultNodeRendering = _doubleArrow;
      EList<KStyleHolder> _renderings = library.getRenderings();
      _xblockexpression = _renderings.add(this.defaultNodeRendering);
    }
    return _xblockexpression;
  }
  
  /**
   * Recursively calls {@link #enrichRendering(EObject)} on this object and all of its children.
   * 
   * @param o object to enrich.
   */
  private void recursivelyEnrichRendering(final EObject o) {
    this.enrichRendering(o);
    EList<EObject> _eContents = o.eContents();
    final Procedure1<EObject> _function = new Procedure1<EObject>() {
      public void apply(final EObject child) {
        KGraphDiagramSynthesis.this.recursivelyEnrichRendering(child);
      }
    };
    IterableExtensions.<EObject>forEach(_eContents, _function);
  }
  
  /**
   * Fallback metchod that does not add any rendering information to the given object.
   * 
   * @param o an EMF object.
   */
  private void _enrichRendering(final EObject o) {
  }
  
  /**
   * Make sure the node has a valid size and a label (if defaults are switched on).
   * 
   * @param node the node whose rendering to enrich.
   */
  private void _enrichRendering(final KNode node) {
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not = (!_defaultsEnabled);
    if (_not) {
      return;
    }
    this.ensureLabel(node);
    final KShapeLayout layoutData = node.<KShapeLayout>getData(KShapeLayout.class);
    boolean _notEquals = (!Objects.equal(layoutData, null));
    if (_notEquals) {
      final EnumSet<SizeConstraint> sizeConstraint = layoutData.<EnumSet<SizeConstraint>>getProperty(LayoutOptions.SIZE_CONSTRAINT);
      boolean _and = false;
      boolean _and_1 = false;
      float _width = layoutData.getWidth();
      boolean _equals = (_width == 0.0f);
      if (!_equals) {
        _and_1 = false;
      } else {
        float _height = layoutData.getHeight();
        boolean _equals_1 = (_height == 0.0f);
        _and_1 = _equals_1;
      }
      if (!_and_1) {
        _and = false;
      } else {
        EnumSet<SizeConstraint> _fixed = SizeConstraint.fixed();
        boolean _equals_2 = sizeConstraint.equals(_fixed);
        _and = _equals_2;
      }
      if (_and) {
        layoutData.setWidth(80);
        layoutData.setHeight(80);
      }
    }
    this._kLibraryExtensions.addRenderingRef(node, this.defaultNodeRendering);
  }
  
  /**
   * Make sure the port has a valid size and a label (if defaults are switched on).
   * 
   * @param port the port whose rendering to enrich.
   */
  private void _enrichRendering(final KPort port) {
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not = (!_defaultsEnabled);
    if (_not) {
      return;
    }
    this.ensureLabel(port);
    final KShapeLayout layoutData = port.<KShapeLayout>getData(KShapeLayout.class);
    boolean _notEquals = (!Objects.equal(layoutData, null));
    if (_notEquals) {
      boolean _and = false;
      float _width = layoutData.getWidth();
      boolean _equals = (_width == 0.0f);
      if (!_equals) {
        _and = false;
      } else {
        float _height = layoutData.getHeight();
        boolean _equals_1 = (_height == 0.0f);
        _and = _equals_1;
      }
      if (_and) {
        layoutData.setWidth(5);
        layoutData.setHeight(5);
      }
    }
  }
  
  /**
   * Possibly adds a proper rendering to the given edge. If no rendering is associated with the edge
   * yet, a default rendering with an arrowhead is added to it.
   * 
   * @param edge the edge whose rendering to enrich.
   */
  private void _enrichRendering(final KEdge edge) {
    boolean _hasRendering = this.hasRendering(edge);
    boolean _not = (!_hasRendering);
    if (_not) {
      KNode _xifexpression = null;
      KNode _target = edge.getTarget();
      KNode _source = edge.getSource();
      boolean _isDescendant = KimlUtil.isDescendant(_target, _source);
      if (_isDescendant) {
        _xifexpression = edge.getSource();
      } else {
        KNode _source_1 = edge.getSource();
        KNode _parent = null;
        if (_source_1!=null) {
          _parent=_source_1.getParent();
        }
        _xifexpression = _parent;
      }
      final KNode parent = _xifexpression;
      KLayoutData _data = null;
      if (parent!=null) {
        _data=parent.<KLayoutData>getData(KLayoutData.class);
      }
      EdgeRouting _property = null;
      if (_data!=null) {
        _property=_data.<EdgeRouting>getProperty(LayoutOptions.EDGE_ROUTING);
      }
      final EdgeRouting routing = _property;
      EList<KGraphData> _data_1 = edge.getData();
      KRenderingRef _createKRenderingRef = KGraphDiagramSynthesis.renderingFactory.createKRenderingRef();
      final Procedure1<KRenderingRef> _function = new Procedure1<KRenderingRef>() {
        public void apply(final KRenderingRef it) {
          boolean _and = false;
          boolean _notEquals = (!Objects.equal(routing, null));
          if (!_notEquals) {
            _and = false;
          } else {
            boolean _equals = Objects.equal(routing, EdgeRouting.SPLINES);
            _and = _equals;
          }
          if (_and) {
            it.setRendering(KGraphDiagramSynthesis.this.defaultSplineRendering);
          } else {
            it.setRendering(KGraphDiagramSynthesis.this.defaultPolylineRendering);
          }
        }
      };
      KRenderingRef _doubleArrow = ObjectExtensions.<KRenderingRef>operator_doubleArrow(_createKRenderingRef, _function);
      _data_1.add(_doubleArrow);
    }
  }
  
  /**
   * Possibly adds a proper rendering to the given label. If no rendering is associated with the label
   * yet, a default text rendering with a predefined font size is added to it. Also, makes sure that
   * an edge label placement is set on the label if it belongs to an edge.
   * 
   * @param edge the edge whose rendering to enrich.
   */
  private void _enrichRendering(final KLabel label) {
    boolean _hasRendering = this.hasRendering(label);
    boolean _not = (!_hasRendering);
    if (_not) {
      KText _createKText = KGraphDiagramSynthesis.renderingFactory.createKText();
      final Procedure1<KText> _function = new Procedure1<KText>() {
        public void apply(final KText text) {
          EList<KGraphData> _data = label.getData();
          _data.add(text);
          KGraphDiagramSynthesis.this._kRenderingExtensions.setFontSize(text, (KlighdConstants.DEFAULT_FONT_SIZE - 2));
          EObject _eContainer = label.eContainer();
          if ((_eContainer instanceof KPort)) {
            KGraphDiagramSynthesis.this._kRenderingExtensions.setFontSize(text, (KlighdConstants.DEFAULT_FONT_SIZE - 3));
          }
        }
      };
      ObjectExtensions.<KText>operator_doubleArrow(_createKText, _function);
    }
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not_1 = (!_defaultsEnabled);
    if (_not_1) {
      return;
    }
    EObject _eContainer = label.eContainer();
    if ((_eContainer instanceof KEdge)) {
      final KLayoutData layoutData = label.<KLayoutData>getData(KLayoutData.class);
      boolean _notEquals = (!Objects.equal(layoutData, null));
      if (_notEquals) {
        final EdgeLabelPlacement edgeLabelPlacement = layoutData.<EdgeLabelPlacement>getProperty(LayoutOptions.EDGE_LABEL_PLACEMENT);
        boolean _equals = Objects.equal(edgeLabelPlacement, EdgeLabelPlacement.UNDEFINED);
        if (_equals) {
          layoutData.<EdgeLabelPlacement>setProperty(
            LayoutOptions.EDGE_LABEL_PLACEMENT, EdgeLabelPlacement.CENTER);
        }
      }
    } else {
      EObject _eContainer_1 = label.eContainer();
      if ((_eContainer_1 instanceof KNode)) {
        EObject _eContainer_2 = label.eContainer();
        final KNode node = ((KNode) _eContainer_2);
        final KLayoutData layoutData_1 = node.<KLayoutData>getData(KLayoutData.class);
        boolean _notEquals_1 = (!Objects.equal(layoutData_1, null));
        if (_notEquals_1) {
          final EnumSet<NodeLabelPlacement> nodeLabelPlacement = layoutData_1.<EnumSet<NodeLabelPlacement>>getProperty(LayoutOptions.NODE_LABEL_PLACEMENT);
          EnumSet<NodeLabelPlacement> _fixed = NodeLabelPlacement.fixed();
          boolean _equals_1 = nodeLabelPlacement.equals(_fixed);
          if (_equals_1) {
            EnumSet<NodeLabelPlacement> _insideCenter = NodeLabelPlacement.insideCenter();
            layoutData_1.<EnumSet<NodeLabelPlacement>>setProperty(
              LayoutOptions.NODE_LABEL_PLACEMENT, _insideCenter);
          }
        }
      }
    }
  }
  
  /**
   * Checks if the given graph element already has a rendering attached.
   * 
   * @param e the element to check.
   * @return {@code true} if the element already has a rendering attached, {@code false} otherwise.
   */
  private boolean hasRendering(final KGraphElement e) {
    EList<KGraphData> _data = e.getData();
    final Function1<KGraphData, Boolean> _function = new Function1<KGraphData, Boolean>() {
      public Boolean apply(final KGraphData it) {
        return Boolean.valueOf((it instanceof KRendering));
      }
    };
    return IterableExtensions.<KGraphData>exists(_data, _function);
  }
  
  /**
   * Ensures that the given element has a label, if possible. If it doesn't, a label is added with its
   * identifier. If there is no identifier, no label is added.
   * 
   * @param e the element to add the label to.
   */
  private void ensureLabel(final KLabeledGraphElement e) {
    boolean _defaultsEnabled = this.defaultsEnabled();
    boolean _not = (!_defaultsEnabled);
    if (_not) {
      return;
    }
    EList<KLabel> _labels = e.getLabels();
    boolean _isEmpty = _labels.isEmpty();
    if (_isEmpty) {
      final KIdentifier identifier = e.<KIdentifier>getData(KIdentifier.class);
      boolean _and = false;
      boolean _notEquals = (!Objects.equal(identifier, null));
      if (!_notEquals) {
        _and = false;
      } else {
        String _id = identifier.getId();
        boolean _isNullOrEmpty = Strings.isNullOrEmpty(_id);
        boolean _not_1 = (!_isNullOrEmpty);
        _and = _not_1;
      }
      if (_and) {
        final KLabel label = KimlUtil.createInitializedLabel(e);
        String _id_1 = identifier.getId();
        label.setText(_id_1);
      }
    }
  }
  
  /**
   * Returns whether or not default stuff should be added or not.
   * 
   * @return {@code true} if default stuff should be added, {@code false} otherwise.
   */
  private boolean defaultsEnabled() {
    boolean _or = false;
    Object _objectValue = this.getObjectValue(KGraphDiagramSynthesis.DEFAULTS);
    boolean _equals = Objects.equal(_objectValue, KGraphDiagramSynthesis.DEFAULTS_ON);
    if (_equals) {
      _or = true;
    } else {
      boolean _and = false;
      Object _objectValue_1 = this.getObjectValue(KGraphDiagramSynthesis.DEFAULTS);
      boolean _equals_1 = Objects.equal(_objectValue_1, KGraphDiagramSynthesis.DEFAULTS_AS_IN_MODEL);
      if (!_equals_1) {
        _and = false;
      } else {
        _and = this.defaults;
      }
      _or = _and;
    }
    return _or;
  }
  
  private void enrichRendering(final EObject edge) {
    if (edge instanceof KEdge) {
      _enrichRendering((KEdge)edge);
      return;
    } else if (edge instanceof KNode) {
      _enrichRendering((KNode)edge);
      return;
    } else if (edge instanceof KPort) {
      _enrichRendering((KPort)edge);
      return;
    } else if (edge instanceof KLabel) {
      _enrichRendering((KLabel)edge);
      return;
    } else if (edge != null) {
      _enrichRendering(edge);
      return;
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(edge).toString());
    }
  }
}
