/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.ptolemy.klighd.PluginConstants;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.ptolemy.moml.ClassType;
import org.ptolemy.moml.EntityType;
import org.ptolemy.moml.PropertyType;
import ptolemy.actor.IOPort;
import ptolemy.actor.TypedCompositeActor;
import ptolemy.actor.parameters.ParameterPort;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.Entity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.NamedObj;
import ptolemy.kernel.util.StringAttribute;
import ptolemy.moml.MoMLParser;
import ptolemy.moml.filter.BackwardCompatibility;

/**
 * Provides an interface to the Ptolemy library to instantiate actors. This is used during the
 * transformation to get our hands at the list of ports defined for an actor.
 * 
 * @author cds
 * @author haf
 * @kieler.rating yellow 2012-06-14 KI-12 cmot, grh
 */
@SuppressWarnings("all")
public class PtolemyInterface {
  /**
   * Accessing annotations.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Labeling nodes and ports.
   */
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * A cache mapping qualified class names of Ptolemy actors to their actual instances. If an actor
   * was already instantiated, there's no need to instantiate it again since that's quite a bit of
   * work.
   */
  private static HashMap<String, Entity> entityCache = new HashMap<String, Entity>();
  
  /**
   * Tries to instantiate the given entity to return a list of its ports. The entity must either be
   * an {@code EntityType} or a {@code ClassType}. If the entity could not be instantiated, an empty
   * list is returned.
   * 
   * @param entity description of the entity.
   * @return list of ports which will be empty if the entity could not be instantiated.
   * @throws Exception if the instantiation fails.
   */
  public List<KPort> getPortsFromImplementation(final EObject entity) {
    ArrayList<KPort> _xblockexpression = null;
    {
      final ArrayList<KPort> result = new ArrayList<KPort>();
      Entity ptActor = null;
      Entity _instantiatePtolemyEntity = this.instantiatePtolemyEntity(entity);
      ptActor = _instantiatePtolemyEntity;
      boolean _notEquals = (!Objects.equal(ptActor, null));
      if (_notEquals) {
        int index = 0;
        List _portList = ptActor.portList();
        for (final Object port : _portList) {
          {
            if ((port instanceof IOPort)) {
              final IOPort ptPort = ((IOPort) port);
              final KPort kPort = KimlUtil.createInitializedPort();
              KLayoutData _layout = this._annotationExtensions.getLayout(kPort);
              _layout.<Integer>setProperty(LayoutOptions.PORT_INDEX, Integer.valueOf(index));
              String _name = ptPort.getName();
              this._labelExtensions.setName(kPort, _name);
              this._markerExtensions.markAsPtolemyElement(kPort);
              List _attributeList = ptPort.attributeList();
              for (final Object attribute : _attributeList) {
                if ((attribute instanceof Attribute)) {
                  this.turnAttributeIntoAnnotation(kPort, ((Attribute) attribute));
                }
              }
              boolean _or = false;
              boolean _or_1 = false;
              boolean _isInput = ptPort.isInput();
              if (_isInput) {
                _or_1 = true;
              } else {
                boolean _hasAnnotation = this._annotationExtensions.hasAnnotation(kPort, "input");
                _or_1 = _hasAnnotation;
              }
              if (_or_1) {
                _or = true;
              } else {
                boolean _hasAnnotation_1 = this._annotationExtensions.hasAnnotation(kPort, "inputoutput");
                _or = _hasAnnotation_1;
              }
              if (_or) {
                this._markerExtensions.markAsInputPort(kPort, true);
              }
              boolean _or_2 = false;
              boolean _or_3 = false;
              boolean _isOutput = ptPort.isOutput();
              if (_isOutput) {
                _or_3 = true;
              } else {
                boolean _hasAnnotation_2 = this._annotationExtensions.hasAnnotation(kPort, "output");
                _or_3 = _hasAnnotation_2;
              }
              if (_or_3) {
                _or_2 = true;
              } else {
                boolean _hasAnnotation_3 = this._annotationExtensions.hasAnnotation(kPort, "inputoutput");
                _or_2 = _hasAnnotation_3;
              }
              if (_or_2) {
                this._markerExtensions.markAsOutputPort(kPort, true);
              }
              boolean _isMultiport = ptPort.isMultiport();
              if (_isMultiport) {
                this._annotationExtensions.addAnnotation(kPort, TransformationConstants.IS_MULTIPORT);
              }
              if ((ptPort instanceof ParameterPort)) {
                this._annotationExtensions.addAnnotation(kPort, TransformationConstants.IS_PARAMETER_PORT);
              } else {
                this._annotationExtensions.addAnnotation(kPort, TransformationConstants.IS_IO_PORT);
              }
              result.add(kPort);
            }
            index = (index + 1);
          }
        }
      }
      _xblockexpression = result;
    }
    return _xblockexpression;
  }
  
  /**
   * Makes an annotation out of the given attribute and attaches it to the given object.
   * Recursively adds attributes of the attributes to the correspondingly created annotations.
   * 
   * @param element the KGraph element to annotate with the transformed attribute.
   * @param ptAttribute the attribute to turn into an annotation.
   */
  private void turnAttributeIntoAnnotation(final EObject element, final Attribute ptAttribute) {
    String _name = ptAttribute.getName();
    String _className = ptAttribute.getClassName();
    final PropertyType property = this._annotationExtensions.addAnnotation(element, _name, "", _className);
    if ((ptAttribute instanceof StringAttribute)) {
      String _valueAsString = ((StringAttribute) ptAttribute).getValueAsString();
      property.setValue(_valueAsString);
    }
    List _attributeList = ptAttribute.attributeList();
    for (final Object attribute : _attributeList) {
      if ((attribute instanceof Attribute)) {
        this.turnAttributeIntoAnnotation(property, ((Attribute) attribute));
      }
    }
  }
  
  /**
   * Tries to instantiate the entity referenced by the given entity type.
   * 
   * @param ptEntity entity type describing the entity to instantiate.
   * @return the instantiated entity.
   * @throws Exception if the instantiation fails.
   */
  protected Entity _instantiatePtolemyEntity(final EntityType ptEntity) {
    String _class1 = ptEntity.getClass1();
    String _name = ptEntity.getName();
    return this.instantiatePtolemyEntityWithCache(_class1, _name);
  }
  
  /**
   * Tries to instantiate the entity referenced by the given class type.
   * 
   * @param ptClass class type describing the entity to instantiate.
   * @return the instantiated entity.
   * @throws Exception if the instantiation fails.
   */
  protected Entity _instantiatePtolemyEntity(final ClassType ptClass) {
    String _extends = ptClass.getExtends();
    String _name = ptClass.getName();
    return this.instantiatePtolemyEntityWithCache(_extends, _name);
  }
  
  /**
   * Tries to instantiate the entity referenced by the given class name.
   * 
   * @param className name of the class of the entity to instantiate.
   * @return the instantiated entity.
   * @throws Exception if the instantiation fails.
   */
  protected Entity _instantiatePtolemyEntity(final String className) {
    return this.instantiatePtolemyEntityWithCache(className, "anEntity");
  }
  
  /**
   * Instantiates the Ptolemy actor with the given class name. The entity name doesn't matter, but
   * makes errors make more sense.
   * 
   * @param className the fully qualified class name of the actor to instantiate.
   * @param entityName the actor's name in the model. Useful for error messages.
   * @return the instantiated entity.
   * @throws CoreException if the actor couldn't be instantiated.
   */
  private Entity instantiatePtolemyEntityWithCache(final String className, final String entityName) {
    try {
      Entity _xblockexpression = null;
      {
        final Entity cachedEntity = PtolemyInterface.entityCache.get(className);
        Entity _xifexpression = null;
        boolean _equals = Objects.equal(cachedEntity, null);
        if (_equals) {
          try {
            boolean _equals_1 = className.equals("ptolemy.domains.modal.kernel.State");
            if (_equals_1) {
              final Entity entity = this.instantiatePtolemyState(className, entityName);
              PtolemyInterface.entityCache.put(className, entity);
              return entity;
            } else {
              final Entity entity_1 = this.instantiatePtolemyActor(className, entityName);
              PtolemyInterface.entityCache.put(className, entity_1);
              return entity_1;
            }
          } catch (final Throwable _t) {
            if (_t instanceof Exception) {
              final Exception e = (Exception)_t;
              String _replace = "Unable to instantiate actor %1 (class \'%2\').".replace("%1", entityName);
              String _replace_1 = _replace.replace("%2", className);
              Status _status = new Status(
                IStatus.WARNING, 
                PluginConstants.PLUGIN_ID, _replace_1, e);
              throw new CoreException(_status);
            } else {
              throw Exceptions.sneakyThrow(_t);
            }
          }
        } else {
          _xifexpression = cachedEntity;
        }
        _xblockexpression = _xifexpression;
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  /**
   * Instantiates a Ptolemy actor of the given class with the given name.
   * 
   * @param className the name of the actor's class.
   * @param entityName the name that should be used for the actor when instantiating it. This has no
   *                   influence on the functionality, but results in more readable error messages if
   *                   anything goes wrong.
   * @return the instantiated actor.
   * @throws Exception if the instantiation fails.
   */
  private Entity instantiatePtolemyActor(final String className, final String entityName) {
    try {
      Entity _xblockexpression = null;
      {
        List _allFilters = BackwardCompatibility.allFilters();
        MoMLParser.setMoMLFilters(_allFilters);
        final MoMLParser parser = new MoMLParser();
        StringConcatenation _builder = new StringConcatenation();
        _builder.append("<entity name=\"TopLevel\" class=\"ptolemy.actor.TypedCompositeActor\">");
        _builder.newLine();
        _builder.append("    ");
        _builder.append("<entity name=\"");
        _builder.append(entityName, "    ");
        _builder.append("\" class=\"");
        _builder.append(className, "    ");
        _builder.append("\" />");
        _builder.newLineIfNotEmpty();
        _builder.append("</entity>");
        _builder.newLine();
        final String xml = _builder.toString();
        String _string = xml.toString();
        final NamedObj parentElement = parser.parse(_string);
        List _entityList = ((TypedCompositeActor) parentElement).entityList();
        Object _get = _entityList.get(0);
        _xblockexpression = ((Entity) _get);
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  /**
   * Instantiates a Ptolemy state entity of the given class with the given name.
   * 
   * @param className the name of the actor's class.
   * @param entityName the name that should be used for the actor when instantiating it. This has no
   *                   influence on the functionality, but results in more readable error messages if
   *                   anything goes wrong.
   * @return the instantiated actor.
   * @throws Exception if the instantiation fails.
   */
  private Entity instantiatePtolemyState(final String className, final String entityName) {
    try {
      Entity _xblockexpression = null;
      {
        List _allFilters = BackwardCompatibility.allFilters();
        MoMLParser.setMoMLFilters(_allFilters);
        final MoMLParser parser = new MoMLParser();
        StringConcatenation _builder = new StringConcatenation();
        _builder.append("<entity name=\"TopLevel\" class=\"ptolemy.domains.modal.modal.ModalController\">");
        _builder.newLine();
        _builder.append("    ");
        _builder.append("<entity name=\"");
        _builder.append(entityName, "    ");
        _builder.append("\" class=\"");
        _builder.append(className, "    ");
        _builder.append("\" />");
        _builder.newLineIfNotEmpty();
        _builder.append("</entity>");
        _builder.newLine();
        final String xml = _builder.toString();
        String _string = xml.toString();
        final NamedObj parentElement = parser.parse(_string);
        List _entityList = ((CompositeEntity) parentElement).entityList();
        Object _get = _entityList.get(0);
        _xblockexpression = ((Entity) _get);
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public Entity instantiatePtolemyEntity(final Object ptClass) {
    if (ptClass instanceof ClassType) {
      return _instantiatePtolemyEntity((ClassType)ptClass);
    } else if (ptClass instanceof EntityType) {
      return _instantiatePtolemyEntity((EntityType)ptClass);
    } else if (ptClass instanceof String) {
      return _instantiatePtolemyEntity((String)ptClass);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(ptClass).toString());
    }
  }
}
