/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation.extensions;

import com.google.common.base.Strings;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.Set;
import org.eclipse.xtext.xbase.lib.Extension;

/**
 * Utility methods used to mark elements by the Ptolemy to KGraph transformation.
 * 
 * @author cds
 * @kieler.rating yellow 2012-07-10 KI-15 cmot, grh
 * 
 * @containsExtensions ptolemy
 */
@SuppressWarnings("all")
public class MarkerExtensions {
  /**
   * We make use of annotations to mark stuff.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Marks the given KGraph element as originating from a Ptolemy model.
   * 
   * @param element the element to mark.
   */
  public void markAsPtolemyElement(final KGraphElement element) {
    this._annotationExtensions.addAnnotation(element, TransformationConstants.ANNOTATION_LANGUAGE, TransformationConstants.ANNOTATION_LANGUAGE_PTOLEMY);
  }
  
  /**
   * Marks the given edge as being undirected. This means that the link's direction has not been
   * inferred yet.
   * 
   * @param edge the edge to mark.
   * @param mark {@code true} if the edge should be marked as being undirected, {@code false} if the
   *             marking should be removed.
   */
  public void markAsUndirected(final KEdge edge, final boolean mark) {
    if (mark) {
      this._annotationExtensions.addAnnotation(edge, "_undirected");
    } else {
      this._annotationExtensions.removeAnnotation(edge, "_undirected");
    }
  }
  
  /**
   * Checks if the given edge is marked as being undirected.
   * 
   * @param edge the edge to check.
   * @return {@code true} if the edge is marked as undirected, {@code false} otherwise.
   */
  public boolean isMarkedAsUndirected(final KEdge edge) {
    return this._annotationExtensions.hasAnnotation(edge, "_undirected");
  }
  
  /**
   * Marks the given port as being an input port.
   * 
   * @param port the port to mark.
   * @param mark {@code true} if the port should be marked, {@code false} if the marking should be
   *             removed.
   */
  public void markAsInputPort(final KGraphElement port, final boolean mark) {
    if (mark) {
      this._annotationExtensions.addAnnotation(port, "_input");
    } else {
      this._annotationExtensions.removeAnnotation(port, "_input");
    }
  }
  
  /**
   * Checks if the given port is marked as being an input port.
   * 
   * @param port the port to check.
   * @return {@code true} if the port is marked as being an input port, {@code false} otherwise.
   */
  public boolean isMarkedAsInputPort(final KGraphElement port) {
    return this._annotationExtensions.hasAnnotation(port, "_input");
  }
  
  /**
   * Marks the given port as being an output port.
   * 
   * @param port the port to mark.
   * @param mark {@code true} if the port should be marked, {@code false} if the marking should be
   *             removed.
   */
  public void markAsOutputPort(final KGraphElement port, final boolean mark) {
    if (mark) {
      this._annotationExtensions.addAnnotation(port, "_output");
    } else {
      this._annotationExtensions.removeAnnotation(port, "_output");
    }
  }
  
  /**
   * Checks if the given port is marked as being an output port.
   * 
   * @param port the port to check.
   * @return {@code true} if the port is marked as being an output port, {@code false} otherwise.
   */
  public boolean isMarkedAsOutputPort(final KGraphElement port) {
    return this._annotationExtensions.hasAnnotation(port, "_output");
  }
  
  /**
   * Marks a node as formerly having been an annotation. Some nodes do not map to actors in the
   * original Ptolemy model, but to annotations. For example, the director used to execute a model
   * is persisted as an annotation in Ptolemy. In the corresponding KGraph model, we need it to be a
   * node to display it correctly.
   * 
   * @param node the node to mark.
   */
  public void markAsFormerAnnotationNode(final KNode node) {
    this._annotationExtensions.addAnnotation(node, "_annotationNode");
  }
  
  /**
   * Checks if the given node is marked as having been created from an annotation.
   * 
   * @param node the node to check.
   * @return {@code true} if it is marked, {@code false} otherwise.
   */
  public boolean isMarkedAsFormerAnnotationNode(final KNode node) {
    return this._annotationExtensions.hasAnnotation(node, "_annotationNode");
  }
  
  /**
   * Marks the given node as being a hypernode.
   * 
   * @param node the node to be marked.
   */
  public void markAsHypernode(final KNode node) {
    KShapeLayout _safeLayout = this._annotationExtensions.getSafeLayout(node);
    _safeLayout.<Boolean>setProperty(LayoutOptions.HYPERNODE, Boolean.valueOf(true));
  }
  
  /**
   * Checks if the given node is marked as being a hypernode.
   * 
   * @param node the node to check.
   * @return {@code true} if the node is marked as being a hypernode, {@code false} otherwise.
   */
  public boolean isMarkedAsHypernode(final KNode node) {
    KShapeLayout _safeLayout = this._annotationExtensions.getSafeLayout(node);
    return (_safeLayout.<Boolean>getProperty(LayoutOptions.HYPERNODE)).booleanValue();
  }
  
  /**
   * Marks the given node as being a comment node.
   * 
   * @param node the node to be marked.
   */
  public void markAsComment(final KNode node) {
    KShapeLayout _safeLayout = this._annotationExtensions.getSafeLayout(node);
    _safeLayout.<Boolean>setProperty(LayoutOptions.COMMENT_BOX, Boolean.valueOf(true));
  }
  
  /**
   * Checks if the given node is marked as being a comment node.
   * 
   * @param node the node to check.
   * @return {@code true} if the node is marked as being a comment node, {@code false} otherwise.
   */
  public boolean isMarkedAsComment(final KNode node) {
    KShapeLayout _safeLayout = this._annotationExtensions.getSafeLayout(node);
    return (_safeLayout.<Boolean>getProperty(LayoutOptions.COMMENT_BOX)).booleanValue();
  }
  
  /**
   * Marks the given node as containing a state machine.
   * 
   * @param node the node to be marked.
   */
  public void markAsStateMachineContainer(final KNode node) {
    this._annotationExtensions.addAnnotation(node, "_stateMachineContainer");
  }
  
  /**
   * Checks if the given node is marked as containing a state machine.
   * 
   * @param node the node to check.
   * @return {@code true} if the node is marked as containing a state machine, {@code false}
   *         otherwise.
   */
  public boolean isMarkedAsStateMachineContainer(final KNode node) {
    return this._annotationExtensions.hasAnnotation(node, "_stateMachineContainer");
  }
  
  /**
   * Marks the given node as being a director node.
   * 
   * @param node the node to be marked.
   */
  public void markAsDirector(final KNode node) {
    this._annotationExtensions.addAnnotation(node, "_director");
  }
  
  /**
   * Checks if the given node is marked as being a director node.
   * 
   * @param node the node to check.
   * @return {@code true} if the node is a director node.
   */
  public boolean isMarkedAsDirector(final KNode node) {
    return this._annotationExtensions.hasAnnotation(node, "_director");
  }
  
  /**
   * Marks the given node as being a parameter node that displays model parameters.
   * 
   * @param node the node to be marked.
   */
  public void markAsParameterNode(final KNode node) {
    this._annotationExtensions.addAnnotation(node, "_parameters");
  }
  
  /**
   * Checks if the given node is marked as being a parameter node.
   * 
   * @param node the node to check.
   * @return {@code true} if the node is a parameter node.
   */
  public boolean isMarkedAsParameterNode(final KNode node) {
    return this._annotationExtensions.hasAnnotation(node, "_parameters");
  }
  
  /**
   * Checks if the given node has a state refinement.
   * 
   * @param node the node to check.
   * @return {@code true} if the node has a refinement.
   */
  public boolean isMarkedAsHavingRefinement(final KNode node) {
    String _annotationValue = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_REFINEMENT_NAME);
    boolean _isNullOrEmpty = Strings.isNullOrEmpty(_annotationValue);
    return (!_isNullOrEmpty);
  }
  
  /**
   * Checks if the given node is a value displaying actor, for instance a Const actor.
   * 
   * @param node the node to check.
   * @return {@code true} if the node represents a value displaying actor.
   */
  public boolean isMarkedAsValueDisplayingActor(final KNode node) {
    String _annotationValue = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    final String propertyValue = Strings.nullToEmpty(_annotationValue);
    Set<String> _keySet = TransformationConstants.VALUE_DISPLAY_MAP.keySet();
    return _keySet.contains(propertyValue);
  }
  
  /**
   * Checks if the given node is a Ptolemy state machine model state.
   * 
   * @param node the node to check.
   * @return {@code true} if the node represents a state.
   */
  public boolean isMarkedAsState(final KNode node) {
    String _annotationValue = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    final String propertyValue = Strings.nullToEmpty(_annotationValue);
    boolean _or = false;
    boolean _equals = propertyValue.equals(TransformationConstants.ENTITY_CLASS_STATE);
    if (_equals) {
      _or = true;
    } else {
      boolean _equals_1 = propertyValue.equals(TransformationConstants.ENTITY_CLASS_FSM_STATE);
      _or = _equals_1;
    }
    return _or;
  }
  
  /**
   * Marks the given node as representing a modal model port by setting its Ptolemy class accordingly.
   * 
   * @param node the node to be marked.
   */
  public void markAsModalModelPort(final KNode node) {
    this._annotationExtensions.addAnnotation(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS, TransformationConstants.PORT_CLASS_REFINEMENT_PORT);
  }
  
  /**
   * Checks if the given node or port is a state machine model port. These ports need to be turned
   * into nodes that can have a special rendering, like in Ptolemy.
   * 
   * @param element the element to check.
   * @return {@code true} if the element represents a modal model port.
   */
  public boolean isMarkedAsModalModelPort(final KGraphElement element) {
    String _annotationValue = this._annotationExtensions.getAnnotationValue(element, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    final String propertyValue = Strings.nullToEmpty(_annotationValue);
    boolean _or = false;
    boolean _equals = propertyValue.equals(TransformationConstants.PORT_CLASS_REFINEMENT_PORT);
    if (_equals) {
      _or = true;
    } else {
      boolean _equals_1 = propertyValue.equals(TransformationConstants.PORT_CLASS_FSM_REFINEMENT_PORT);
      _or = _equals_1;
    }
    return _or;
  }
}
